from abc import abstractmethod
from decimal import Decimal
from typing import Any, Union

from loguru import logger

from otlpy.base.net import AsyncHttpClient
from otlpy.base.order import (
    ORDER_TYPE,
    Buy,
    Cancel,
    Order,
    OrderAPI2,
    Replace,
    Sell,
)
from otlpy.kis.common import Common


def str_order_type(order_type: ORDER_TYPE, /) -> str:
    if order_type == ORDER_TYPE.LIMIT:
        s = "01"
    elif order_type == ORDER_TYPE.MARKET:
        s = "02"
    else:
        assert False
    return s


class DomesticFutureOption(OrderAPI2):
    def __init__(self, common: Common, client: AsyncHttpClient, /) -> None:
        self.__common = common
        self.__client = client

    @property
    def _common(self, /) -> Common:
        return self.__common

    @property
    def _client(self, /) -> AsyncHttpClient:
        return self.__client

    async def token(self, /) -> None:
        await self._common.token(self._client)

    async def _new_order(
        self,
        order: Union[Buy, Sell],
        /,
        *,
        omsg: str,
    ) -> None:
        common = self._common
        client = self._client
        otype = str_order_type(order.otype)
        tr_id = "TTTO1101U"
        url_path = "/uapi/domestic-futureoption/v1/trading/order"
        data = {
            "ORD_PRCS_DVSN_CD": "02",
            "CANO": common.account_cano_domestic_futureoption,
            "ACNT_PRDT_CD": common.account_prdt_domestic_futureoption,
            "SLL_BUY_DVSN_CD": omsg,
            "SHTN_PDNO": order.ticker,
            "ORD_QTY": f"{order.qty:f}",
            "UNIT_PRICE": f"{order.price:f}",
            "NMPR_TYPE_CD": otype,
            "KRX_NMPR_CNDT_CD": "0",
            "ORD_DVSN_CD": otype,
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "custtype": common.account_custtype,
            "hashkey": await common.hash(client, data),
        }
        err, rheaders, rdata = await client.post(url_path, headers, data)
        if err:
            return
        if rdata["rt_cd"] != "0":
            logger.error(
                f"\n, {url_path}"
                f"\n, {headers}"
                f"\n, {data}"
                f"\n, {rheaders}"
                f"\n, {rdata}"
            )
            return
        order.acknowledgment(rdata, rdata["output"]["ODNO"], order.qty)

    async def _cancel_or_replace_order(
        self,
        order: Union[Cancel, Replace],
        /,
        *,
        omsg: str,
        ack_qty: Decimal,
    ) -> None:
        common = self._common
        client = self._client
        rdata1 = order.origin.rdata["output"]
        otype = str_order_type(order.otype)
        tr_id = "TTTO1103U"
        url_path = "/uapi/domestic-futureoption/v1/trading/order-rvsecncl"
        data = {
            "ORD_PRCS_DVSN_CD": "02",
            "CANO": common.account_cano_domestic_futureoption,
            "ACNT_PRDT_CD": common.account_prdt_domestic_futureoption,
            "RVSE_CNCL_DVSN_CD": omsg,
            "ORGN_ODNO": rdata1["ODNO"],
            "ORD_QTY": "",
            "UNIT_PRICE": f"{order.price:f}",
            "NMPR_TYPE_CD": otype,
            "KRX_NMPR_CNDT_CD": "0",
            "RMN_QTY_YN": "Y",
            "ORD_DVSN_CD": otype,
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "custtype": common.account_custtype,
            "hashkey": await common.hash(client, data),
        }
        err, rheaders, rdata = await client.post(url_path, headers, data)
        if err:
            return
        if rdata["rt_cd"] != "0":
            logger.error(
                f"\n, {url_path}"
                f"\n, {headers}"
                f"\n, {data}"
                f"\n, {rheaders}"
                f"\n, {rdata}"
            )
            return
        order.acknowledgment(rdata, rdata["output"]["ODNO"], ack_qty)

    async def buy(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
        *,
        comment: str = "",
    ) -> Buy:
        order = Buy(order_type, ticker, qty, price, comment=comment)
        await self._new_order(order, omsg="02")
        return order

    async def sell(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
        *,
        comment: str = "",
    ) -> Sell:
        order = Sell(order_type, ticker, qty, price, comment=comment)
        await self._new_order(order, omsg="01")
        return order

    async def cancel(
        self,
        origin: Order,
        /,
        *,
        comment: str = "",
    ) -> Cancel:
        order = Cancel(origin, ORDER_TYPE.LIMIT, comment=comment)
        await self._cancel_or_replace_order(
            order, omsg="02", ack_qty=Decimal()
        )
        return order

    async def replace(
        self,
        origin: Order,
        order_type: ORDER_TYPE,
        /,
        price: Decimal,
        *,
        comment: str = "",
    ) -> Replace:
        order = Replace(origin, order_type, price, comment=comment)
        await self._cancel_or_replace_order(
            order, omsg="01", ack_qty=order.qty
        )
        return order

    async def all_orders(
        self,
        yyyymmdd: str,
        /,
        *,
        tr_cont: str,
        ctx_area_fk200: str,
        ctx_area_nk200: str,
        ticker: str = "",
    ) -> tuple[bool, Any]:
        common = self._common
        tr_id = "TTTO5201R"
        url_path = "/uapi/domestic-futureoption/v1/trading/inquire-ccnl"
        data = {
            "CANO": common.account_cano_domestic_futureoption,
            "ACNT_PRDT_CD": common.account_prdt_domestic_futureoption,
            "STRT_ORD_DT": yyyymmdd,
            "END_ORD_DT": yyyymmdd,
            "SLL_BUY_DVSN_CD": "00",
            "CCLD_NCCS_DVSN": "00",
            "SORT_SQN": "DS",
            "STRT_ODNO": "",
            "PDNO": ticker,
            "MKET_ID_CD": "",
            "CTX_AREA_FK200": ctx_area_fk200,
            "CTX_AREA_NK200": ctx_area_nk200,
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "custtype": common.account_custtype,
            "tr_cont": tr_cont,
        }
        err, rheaders, rdata = await self._client.get(url_path, headers, data)
        if err:
            return False, None
        if rdata["rt_cd"] != "0":
            logger.error(
                f"\n, {url_path}"
                f"\n, {headers}"
                f"\n, {data}"
                f"\n, {rheaders}"
                f"\n, {rdata}"
            )
            return False, None
        if rheaders["tr_cont"] == "F" or rheaders["tr_cont"] == "M":
            return True, rdata
        if rheaders["tr_cont"] == "D" or rheaders["tr_cont"] == "E":
            return False, rdata
        assert False

    async def limitorderbook(self, ticker: str, /) -> Any:
        meta = self.meta()
        common = self._common
        tr_id = "FHMIF10010000"
        url_path = (
            "/uapi/domestic-futureoption/v1/quotations/inquire-asking-price"
        )
        data = {
            "FID_COND_MRKT_DIV_CODE": meta["fid_cond_mrkt_div_code"],
            "FID_INPUT_ISCD": ticker,
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "custtype": common.account_custtype,
        }
        err, rheaders, rdata = await self._client.get(url_path, headers, data)
        if err:
            return None
        if rdata["rt_cd"] != "0":
            logger.error(
                f"\n, {url_path}"
                f"\n, {headers}"
                f"\n, {data}"
                f"\n, {rheaders}"
                f"\n, {rdata}"
            )
            return None
        return rdata

    @abstractmethod
    def meta(self, /) -> dict[str, str]:
        raise NotImplementedError


class IndexFuture(DomesticFutureOption):
    def meta(self, /) -> dict[str, str]:
        return {
            "fid_cond_mrkt_div_code": "F",
        }


class IndexOption(DomesticFutureOption):
    def meta(self, /) -> dict[str, str]:
        return {
            "fid_cond_mrkt_div_code": "O",
        }


class StockFuture(DomesticFutureOption):
    def meta(self, /) -> dict[str, str]:
        return {
            "fid_cond_mrkt_div_code": "JF",
        }


class StockOption(DomesticFutureOption):
    def meta(self, /) -> dict[str, str]:
        return {
            "fid_cond_mrkt_div_code": "JO",
        }
